// C Source File
// Created 02/08/2004; 12:52:21

#include "game.h"

void setGravity (unsigned short g) {
	Gravity = g>16?16:g;
}

void vectAdd (Vect * v1, Vect * v2, Vect * v3) {
	short dir1, dir2, norm1, norm2;

	if (v1->norm || v2->norm) {
		if (v1->norm == 0) {
			v3->dir = v2->dir;
			v3->norm = v2->norm;
			return;
		}
		if (v2->norm == 0) {
			v3->dir = v1->dir;
			v3->norm = v1->norm;
			return;
		}
		if (((v1->dir + (NBDIR/2)) %NBDIR) == v2->dir) {
			if (v1->norm < v2->norm) {
				v3->dir = v2->dir;
				v3->norm = v2->norm - v1->norm;
			} else {
				v3->dir = v1->dir;
				v3->norm = v1->norm - v2->norm;
			}
		}
	
		if (v1->dir < v2->dir) {
			dir1 = v1->dir;
			norm1 = v1->norm;
			dir2 = v2->dir;
			norm2 = v2->norm;
		} else {
			dir1 = v2->dir;
			norm1 = v2->norm;
			dir2 = v1->dir;		
			norm2 = v1->norm;
		}
		if (((dir1 + NBDIR) - dir2) < (dir2 - dir1))
			dir1 += NBDIR;

		v3->dir = (dir1 * norm1 + dir2 * norm2)/(norm1 + norm2);
		v3->norm = CosTable[(dir1 - dir2 + NBDIR) % NBDIR] * (norm1 + norm2);
	} else {
		v3->dir = 0;
		v3->norm = 0;
	}
}

void translate (short * x, short * y, Vect * v) {
	(*x) += (((v->norm) >> TRANSLATE_DIV) * SinTable[v->dir]) >> SINCOS;
	(*y) += (((v->norm) >> TRANSLATE_DIV) * CosTable[v->dir]) >> SINCOS;
}

void collide (Vect * inertia, short obstacle_orientation, Vect * extForce) {
	extForce->norm = COLL_INCR + inertia->norm;
	extForce->dir = obstacle_orientation;
	
	inertia->norm = (inertia->norm * COLL_DECR) / 32;
}

void collide_ship (Vect * inertia, short obstacle_orientation, Vect * extForce, short norm) {
	extForce->norm = COLL_INCR + norm;
	extForce->dir = obstacle_orientation;
	
	inertia->norm = (inertia->norm * COLL_DECR) / 32;
}

void thrustToInertia(Vect * v1, Vect * v2, unsigned char slide, short * dirs) {
	short i;
	short ineInc = 0, thrust;
	thrust = (CosTable[(v1->dir - v2->dir + NBDIR) % NBDIR] * v1->norm);
	if (thrust > v2->norm)
		ineInc = thrust - v2->norm;
	if (slide) {
		if (v1->norm) {
			for (i=0; i<SLIDE_GRIP-1; i++)
				dirs[i] = dirs[i+1];
			dirs[SLIDE_GRIP-1] = v1->dir;
			v2->dir = dirs[0];
		} else {
			for (i=0; i<SLIDE_GRIP; i++)
				dirs[i] = v1->dir;
		}
	} else {
		for (i=0; i<SLIDE_GRIP; i++)
			dirs[i] = v1->dir;
		v2->dir = v1->dir;
	}
	v2->norm += ineInc;
}
