// C Source File
// Created 18/04/2005; 16:15:49

#include "Mode7Engine.h"

// Compress Part

typedef struct node_ {
	unsigned short nbocc;
	unsigned char isleaf;
	unsigned char l_id;
	unsigned char r_id;
	struct node_ * l_node;
	struct node_ * r_node;
} node_;

typedef struct leaf {
	unsigned short nbocc;
	unsigned char isleaf;
	unsigned char id;
} leaf;

typedef struct code {
	unsigned char nbbit;
	unsigned char bits[16];
} code;

static unsigned char * comp;
static unsigned char numbit;

CALLBACK static inline short leaf_comp(const void *a, const void *b) {
	return ((leaf*)a)->nbocc - ((leaf*)b)->nbocc;
}

static inline code lbit(code cod) {
	cod.bits[cod.nbbit >> 3] &= ~(0x80 >> (((cod.nbbit)++) & 0x7));
	return cod;
}

static inline code rbit_(code cod) {
	cod.bits[cod.nbbit >> 3] |= (0x80 >> (((cod.nbbit)++) & 0x7));
	return cod;
}

static inline void wbit(unsigned char bit) {
	if(bit) *comp |= (0x80 >> numbit);
	else  *comp &= ~(0x80 >> numbit);
	if(!((++numbit) & 0x7)) {++comp;numbit=0;}
}

static inline void wid(unsigned char id) {
	unsigned short i;
	for(i=8;i--;)
		wbit(id & 0x80), id<<=1;
}

static inline void wcode(code cod) {
	unsigned short i;
	for(i=0; i<cod.nbbit; i++)
		wbit(cod.bits[i>>3] & 0x80>>(i&0x7));
}

static inline void btree_read(node_ * tree, code * codes, code cod) {
	wbit(0);
	if(!(tree->l_node)) codes[tree->l_id] = lbit(cod), wbit(1), wid(tree->l_id);
	else btree_read(tree->l_node, codes, lbit(cod));
	if(!(tree->r_node)) codes[tree->r_id] = rbit_(cod), wbit(1), wid(tree->r_id);
	else btree_read(tree->r_node, codes, rbit_(cod));
}

unsigned char * compress(unsigned char * data, unsigned short size, unsigned short * sizeComp) {
	unsigned char * dataComp;
	unsigned short nbleaf;
	code codes[256];
	leaf leafs[256];
	leaf *tableaf;
	unsigned short i;

	for(i=256;i--;) codes[i].nbbit = 0;

	for(i=256;i--;) {
		leafs[i].nbocc = 0;
		leafs[i].id = i;
		leafs[i].isleaf = 1;
	}

	for(i=size; i--;) leafs[data[i]].nbocc++;
		
	qsort(leafs, 256, sizeof(leaf), leaf_comp);
	for(i=256; i-- && leafs[i].nbocc;);
	tableaf = &leafs[256];
	nbleaf = 255-i;

	node_ * nodes;
	node_ ** nodesptrs;

	if(nbleaf!=1) {

		nodes = (node_*)memset(Malloc((nbleaf-1) * sizeof(node_)), 0, (nbleaf-1) * sizeof(node_)) + nbleaf-1;
	
		nodesptrs = (node_ **)Malloc(sizeof(node_ *) * nbleaf);
	
		for(i=nbleaf; i--;)
			*(nodesptrs++) = (node_*)(--tableaf);
	
		nodesptrs -= nbleaf;
	
		for(i = nbleaf-1; i--;) {
			(*(node_*)(--nodes)).nbocc = (*(node_**)(nodesptrs+i))->nbocc + (*(node_**)(nodesptrs+i+1))->nbocc;
			
			if((*(node_**)(nodesptrs+i))->isleaf) (*(node_*)(nodes)).l_id = (*(leaf**)(nodesptrs+i))->id;
			else (*(node_*)(nodes)).l_node = *(node_**)(nodesptrs+i);
			if((*(node_**)(nodesptrs+i+1))->isleaf) (*(node_*)(nodes)).r_id = (*(leaf**)(nodesptrs+i+1))->id;
			else (*(node_*)(nodes)).r_node = *(node_**)(nodesptrs+i+1);
			unsigned short j;
			for(j=i;j--;) if((*(node_*)(nodes)).nbocc <= (*(node_**)(nodesptrs+j))->nbocc) break;
			j++;
			
			memmove(nodesptrs+j+1,nodesptrs+j,(i-j)*sizeof(void*));
			*(node_**)(nodesptrs+j) = nodes;
		}
		Free(nodesptrs);

	} else {
		nodes = (node_*)memset(Malloc(sizeof(node_)), 0, sizeof(node_));
		nodes->l_id = leafs[255].id;
		nbleaf++;
		tableaf--;
	}

	unsigned short offset = ((nbleaf<<1)-1) + (nbleaf<<3) + 16;
	*sizeComp = (offset>>3) + ((offset & 0x7)?1:0);
	dataComp = (unsigned char *)Malloc(sizeof(unsigned char) * (*sizeComp));
	comp = dataComp+2;
	code cod;
	cod.nbbit = 0;
	numbit = 0;
	btree_read(nodes, codes, cod);
	
	unsigned long sum = 0;
	
	for(i=nbleaf; i--;)
		sum += tableaf->nbocc * codes[tableaf->id].nbbit, tableaf++;

	*sizeComp = ((sum+offset)>>3) + (((sum+offset) & 0x7)?1:0);
	*(unsigned short*)dataComp = size;
	comp -= (long)dataComp;
	dataComp = (unsigned char *)Realloc(dataComp, sizeof(unsigned char) * (*sizeComp));
	comp += (long)dataComp;
	
	for(i=0; i<size; i++) wcode(codes[data[i]]);
		
	Free(nodes);
	return dataComp;
}

// Uncompress Part

typedef struct node {
	unsigned char l_id;
	unsigned char r_id;
	struct node * l_node;
	struct node * r_node;
} node;

static node * nod;

static inline unsigned char rbit(void) {
	if(!((++numbit) & 0x7)) {++comp;numbit=0;}
	return *comp & (0x80 >> numbit);
}

static inline unsigned char rid(void) {
	unsigned short i;
	unsigned char id=0;
	for(i=8;i--;)
		id <<= 1, id |= (rbit()?1:0);
	return id;
}

static inline node * mtree_read(unsigned char * id) {
	node * n = nod;
	if(rbit()) { *id = rid(); return NULL;	}
	nod++;
	n->l_node = mtree_read(&(n->l_id));
	n->r_node = mtree_read(&(n->r_id));
	return n;
}

static inline unsigned char rcode(node * n) {
	if(!rbit()) {
		if(!(n->l_node)) return n->l_id;
		else return rcode(n->l_node);
	}
	if(!(n->r_node)) return n->r_id;
	else return rcode(n->r_node);
}

unsigned char * uncompress(unsigned char * dataComp, unsigned short * size) {
	unsigned char * data;
	node nodes[255];
	nod = nodes;
	comp = dataComp+1;
	numbit = 7;
	mtree_read(NULL);
	*size = *(unsigned short*)dataComp;
	data = (unsigned char *)Malloc(sizeof(unsigned char) * (*size));
	unsigned short i;
	for(i=0;i<(*size);i++)
		data[i] = rcode(nodes);
	return data;
}
